/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   ssvn_job.c
 *    ^$JOB ssv
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <ctype.h>

#include "mpsdef.h"
#include "transact.h"
#include "namespace.h"
#include "freem.h"
#include "mref.h"


void ssvn_job_update(void)
{

    char k_buf[512];
    char d_buf[512];

#ifndef AMIGA68K
    uid_t uid;
    gid_t gid;

    uid = getuid ();
    gid = getgid ();
#else
    int uid = 0;
    int gid = 0;
#endif

    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202CHARACTER\201", pid);
    global (set_sym, k_buf, "M\201");

    snprintf (d_buf, 512 - 1, "%s\201", curdir);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202CWD\201", pid);
    global (set_sym, k_buf, d_buf);
    
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202GLOBAL\201", pid);
    global (set_sym, k_buf, " \201");

    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202LOCK\201", pid);
    global (set_sym, k_buf, " \201");

    snprintf (d_buf, 512 - 1, "%s\201", oucpath);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202IOPATH\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202PRIORITY\201", pid);
    snprintf (d_buf, 512 - 1, "%d\201", getpriority (PRIO_PROCESS, pid));
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%s\201", rou_name);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202ROUTINE\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "0:\"%s\"\201", dev[HOME]);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202$PRINCIPAL\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%d\201", tp_level);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202$TLEVEL\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "0:\"%s\"\201", dev[HOME]);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202$IO\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%d\201", uid);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202USER\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%d\201", gid);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202GROUP\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%s\201", nsname);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202NAMESPACE\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%d\201", pid);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202PID\201", pid);
    global (set_sym, k_buf, d_buf);

    switch (fp_mode) {
        
        case 0:

            snprintf (d_buf, 512 - 1, "FIXED\201");
            break;

        default:

            snprintf (d_buf, 512 - 1, "IEEE754\201");
            snprintf (fp_conversion, 9, "%%.%df\201", DBL_DIG);
            zprecise = DBL_DIG;
            
            break;

    }

    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202MATH\201", pid);
    global (set_sym, k_buf, d_buf);

    snprintf (d_buf, 512 - 1, "%d\201", en_revstrf);
    snprintf (k_buf, 512 - 1, "^$JOB\202%d\202ZREVSTR\201", pid);
    global (set_sym, k_buf, d_buf);
    
    stcpy (d_buf, rou_name);
    stcnv_m2c (d_buf);

#if !defined(__FreeBSD__)
    sprintf (k_buf, "    freem: namespace %s routine %s", nsname, d_buf);
    strcpy (m_argv[0], k_buf);
#endif

}


void ssvn_job(short action, char *key, char *data)
{
    int errsav;
    pid_t t_pid;
    int i;
    
    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));
    char *kbuf = (char *) malloc (STRLEN * sizeof(char));
    char *tbuf = (char *) malloc (STRLEN * sizeof(char));

    stcpy (kbuf, key);

    mref_init (r, MREF_RT_SSVN, "^$JOB");
    internal_to_mref (r, kbuf);

    t_pid = atol (mref_get_subscript (r, 0));

    if (r->subscript_count > 1) {
        for (i = 0; i < strlen (r->subscripts[1]); i++) {
            r->subscripts[1][i] = toupper (r->subscripts[1][i]);
        }
    }

    key = mref_to_internal (r);
    stcpy (kbuf, key);

    switch (action) {
    
        case set_sym:

            {
                char cbuf[STRLEN];

                snprintf (cbuf, STRLEN - 1, "^$JOB\202%d\202NAMESPACE\201", pid);

                if (stcmp (kbuf, cbuf) == 0) {
                    
                    stcpy (tbuf, data);
                    stcnv_m2c (tbuf);

                    set_namespace (tbuf, FALSE);

                    ierr = OK;
                    return;

                }

                if (strcmp (mref_get_subscript (r, 1), "CWD") == 0) {

                    stcpy (tbuf, data);
                    stcnv_m2c (tbuf);

                    if (chdir (tbuf) != 0) {
                        ierr = PROTECT;
                        return;
                    }
                    else {
                        strcpy (curdir, tbuf);
                        global (set_sym, kbuf, data);

                        ierr = OK;
                        return;
                    }

                }
                else if (strcmp (mref_get_subscript (r, 1), "EVENT") == 0) {
                    global (set_sym, kbuf, data);

                    ierr = OK;
                    return;
                } 
                else if (strcmp (mref_get_subscript (r, 1), "PRIORITY") == 0) {

                    int new_pri;                   
                    
                    stcpy (tbuf, data);
                    stcnv_m2c (tbuf);

                    new_pri = atoi (tbuf);

                    
                    i = nice (40);
                    nice ((int) (new_pri - 40));

                    global (set_sym, kbuf, data);

                    ierr = OK;
                    return;

                }
                else if (strcmp (mref_get_subscript (r, 1), "IOPATH") == 0) {

                    stcpy (oucpath, data);
                    global (set_sym, kbuf, data);

                    ierr = OK;
                    return;
                    
                }
                else if (strcmp (mref_get_subscript (r, 1), "ZREVSTR") == 0) {

                    short tres = tvexpr (data);

                    if (tres == TRUE) {
                        en_revstrf = 1;
                    }
                    else {
                        en_revstrf = 0;
                    }                                        
                    
                    global (set_sym, kbuf, data);

                    ierr = OK;
                    return;
                    
                }
                else if (strcmp (mref_get_subscript (r, 1), "MATH") == 0) {

                    if (stcmp (data, "IEEE754\201") == 0) {

                        fp_mode = 1;
                        global (set_sym, kbuf, data);

                        snprintf (fp_conversion, 9, "%%.%df\201", DBL_DIG);
                        zprecise = DBL_DIG;

                        ierr = OK;
                        return;
                        
                    }
#if !defined(__arm__)                    
                    else if (stcmp (data, "FIXED\201") == 0) {

                        fp_mode = 0;
                        global (set_sym, kbuf, data);
                        
                        zprecise = 100;
                        
                        ierr = OK;
                        return;
                        
                    }
#endif                    
                    else {
                        ierr = M29;
                        return;
                    }
                    
                }                

                ierr = M29;
                return;
            }

    	case get_sym:
    	case dat:
    	case order:
    	case query:
    	case bigquery:
    	case getnext:
    	case m_alias:
    	case zdata:
    	
        	global (action, kbuf, data);

            /* here, we need to convert the data on the way out too */                                    
    		if ((action == query) && (stlen (data) > 0)) {
                if (strncmp (data, "^$JOB", 5) != 0) {
                    snprintf (kbuf, STRLEN - 1, "^$JOB%s\201", data);
                    stcpy (data, kbuf);
                }
    		}

    		ierr = OK;
    		return;
    	
        case killone:
    	case kill_sym:
    		
            /* check if t_pid exists */
            if (kill (t_pid, 0) == 0) {
           
                /* send SIGTERM to it, or produce an error message */
                if (kill (t_pid, SIGTERM) == -1) {                
                    
                    errsav = errno;

                    set_io (UNIX);
                    fprintf (stderr, "freem:  error sending SIGTERM to job ID %d (%s)\n", t_pid, strerror (errsav));
                    set_io (MUMPS);

                }

            }
            else {        
                /* t_pid doesn't exist; just kill the global node directly */
                global (action, kbuf, data);
            }

    		return;


    	default:
    		ierr = INVREF;
    		return;

    }

    *data = EOL;
    return;
}
